@@ -9,8 +9,8 @@
 
 from i18n import _
 import changelog, byterange, url, error
-import localrepo, manifest, util, store
-import urllib, urllib2, errno
+import localrepo, manifest, util, scmutil, store
+import urllib, urllib2, errno, os
 
 class httprangereader(object):
     def __init__(self, url, opener):
@@ -26,20 +26,17 @@ class httprangereader(object):
         end = ''
         if bytes:
             end = self.pos + bytes - 1
-        req.add_header('Range', 'bytes=%d-%s' % (self.pos, end))
+        if self.pos or end:
+            req.add_header('Range', 'bytes=%d-%s' % (self.pos, end))
 
         try:
             f = self.opener.open(req)
             data = f.read()
-            if hasattr(f, 'getcode'):
-                # python 2.6+
-                code = f.getcode()
-            elif hasattr(f, 'code'):
-                # undocumented attribute, seems to be set in 2.4 and 2.5
-                code = f.code
-            else:
-                # Don't know how to check, hope for the best.
-                code = 206
+            # Python 2.6+ defines a getcode() function, and 2.4 and
+            # 2.5 appear to always have an undocumented code attribute
+            # set. If we can't read either of those, fall back to 206
+            # and hope for the best.
+            code = getattr(f, 'getcode', lambda : getattr(f, 'code', 206))()
         except urllib2.HTTPError, inst:
             num = inst.code == 404 and errno.ENOENT or None
             raise IOError(num, inst)
@@ -67,17 +64,29 @@ def build_opener(ui, authinfo):
     urlopener = url.opener(ui, authinfo)
     urlopener.add_handler(byterange.HTTPRangeHandler())
 
-    def opener(base):
-        """return a function that opens files over http"""
-        p = base
-        def o(path, mode="r", atomictemp=None):
-            if 'a' in mode or 'w' in mode:
-                raise IOError('Permission denied')
-            f = "/".join((p, urllib.quote(path)))
-            return httprangereader(f, urlopener)
-        return o
+    class statichttpvfs(scmutil.abstractvfs):
+        def __init__(self, base):
+            self.base = base
 
-    return opener
+        def __call__(self, path, mode="r", atomictemp=None):
+            if mode not in ('r', 'rb'):
+                raise IOError('Permission denied')
+            f = "/".join((self.base, urllib.quote(path)))
+            return httprangereader(f, urlopener)
+
+        def join(self, path):
+            if path:
+                return os.path.join(self.base, path)
+            else:
+                return self.base
+
+    return statichttpvfs
+
+class statichttppeer(localrepo.localpeer):
+    def local(self):
+        return None
+    def canpush(self):
+        return False
 
 class statichttprepository(localrepo.localrepository):
     def __init__(self, ui, path):
@@ -85,17 +94,21 @@ class statichttprepository(localrepo.localrepository):
         self.ui = ui
 
         self.root = path
-        self.path, authinfo = url.getauthinfo(path.rstrip('/') + "/.hg")
+        u = util.url(path.rstrip('/') + "/.hg")
+        self.path, authinfo = u.authinfo()
 
         opener = build_opener(ui, authinfo)
         self.opener = opener(self.path)
+        self.vfs = self.opener
+        self._phasedefaults = []
 
-        # find requirements
         try:
-            requirements = self.opener("requires").read().splitlines()
+            requirements = scmutil.readrequires(self.opener, self.supported)
         except IOError, inst:
             if inst.errno != errno.ENOENT:
                 raise
+            requirements = set()
+
             # check if it is a non-empty old-style repository
             try:
                 fp = self.opener("00changelog.i")
@@ -107,29 +120,26 @@ class statichttprepository(localrepo.localrepository):
                 # we do not care about empty old-style repositories here
                 msg = _("'%s' does not appear to be an hg repository") % path
                 raise error.RepoError(msg)
-            requirements = []
-
-        # check them
-        for r in requirements:
-            if r not in self.supported:
-                raise error.RequirementError(
-                        _("requirement '%s' not supported") % r)
 
         # setup store
         self.store = store.store(requirements, self.path, opener)
         self.spath = self.store.path
         self.sopener = self.store.opener
+        self.svfs = self.sopener
         self.sjoin = self.store.join
+        self._filecache = {}
+        self.requirements = requirements
 
         self.manifest = manifest.manifest(self.sopener)
         self.changelog = changelog.changelog(self.sopener)
         self._tags = None
         self.nodetagscache = None
-        self._branchcache = None
-        self._branchcachetip = None
+        self._branchcaches = {}
         self.encodepats = None
         self.decodepats = None
-        self.capabilities = self.capabilities.difference(["pushkey"])
+
+    def _restrictcapabilities(self, caps):
+        return caps.difference(["pushkey"])
 
     def url(self):
         return self._url
@@ -137,6 +147,9 @@ class statichttprepository(localrepo.localrepository):
     def local(self):
         return False
 
+    def peer(self):
+        return statichttppeer(self)
+
     def lock(self, wait=True):
         raise util.Abort(_('cannot lock static-http repository'))
 
